@@ -1,5 +1,5 @@
 /* pwcheck.c -- Unix pwcheck daemon
-   $Id: pwcheck.c,v 1.5 2000/03/09 01:39:44 leg Exp $
+   $Id: pwcheck.c,v 1.6 2000/05/22 20:46:37 leg Exp $
 Copyright 1998, 1999 Carnegie Mellon University
 
                       All Rights Reserved
@@ -22,17 +22,19 @@ ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 ******************************************************************/
 
+#include <unistd.h>
 #include <stdio.h>
 #include <errno.h>
 #include <sys/types.h>
+#include <sys/param.h>
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <sys/uio.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 #ifdef HAVE_PATHS_H
 #include <paths.h>
 #endif
-#include <unistd.h>
 #include <syslog.h>
 
 #if !defined(_PATH_PWCHECKPID)
@@ -43,12 +45,10 @@ OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 #endif
 #endif
 
-#include <config.h>
-
-extern int errno;
+#include "config.h"
 
 void newclient(int);
-int retry_write(int, const char *, unsigned);
+int retry_write(int, const char *, unsigned int);
 
 /*
  * Unix pwcheck daemon-authenticated login (shadow password)
@@ -57,9 +57,11 @@ int retry_write(int, const char *, unsigned);
 int
 main()
 {
-    char fnamebuf[1024];
+    char fnamebuf[MAXPATHLEN];
     int s;
     int c;
+    int count;
+    int rc;
     struct sockaddr_un srvaddr;
     struct sockaddr_un clientaddr;
     int r;
@@ -69,6 +71,61 @@ main()
     FILE *fp = NULL;
     pid_t pid;
 
+    openlog("pwcheck", LOG_NDELAY, LOG_AUTH);
+
+    /* Daemonize. */
+    count = 5;
+    while (count--) {
+	pid = fork();
+            
+	if (pid > 0)
+	    _exit(0);               /* parent dies */
+            
+	if ((pid == -1) && (errno == EAGAIN)) {
+	    syslog(LOG_WARNING, "master fork failed (sleeping): %m");
+	    sleep(5);
+	    continue;
+	}
+    }
+    if (pid == -1) {
+	rc = errno;
+	syslog(LOG_ERR, "FATAL: master fork failed: %m");
+	fprintf(stderr, "pwcheck: ");
+	errno = rc;
+	perror("fork");
+	exit(1);
+    }
+
+    /*
+     * We're now running in the child. Lose our controlling terminal
+     * and obtain a new process group.
+     */
+    if (setsid() == -1) {
+	rc = errno;
+	syslog(LOG_ERR, "FATAL: setsid: %m");
+	fprintf(stderr, "pwcheck: ");
+	errno = rc;
+	perror("setsid");
+	exit(1);
+    }
+        
+    s = open("/dev/null", O_RDWR, 0);
+    if (s == -1) {
+	rc = errno;
+	syslog(LOG_ERR, "FATAL: /dev/null: %m");
+	fprintf(stderr, "pwcheck: ");
+	errno = rc;
+	perror("/dev/null");
+	exit(1);
+            
+    }
+    dup2(s, fileno(stdin));
+    dup2(s, fileno(stdout));
+    dup2(s, fileno(stderr));
+    if (s > 2) {
+	close(s);
+    }
+
     /*
      *   Record process ID - shamelessly stolen from inetd (I.V.)
      */
@@ -89,21 +146,24 @@ main()
 	exit(1);
     }
 
-    strcpy(fnamebuf, PWCHECKDIR);
-    strcat(fnamebuf, "/pwcheck");
+    strncpy(fnamebuf, PWCHECKDIR, sizeof(fnamebuf));
+    strncpy(fnamebuf + sizeof(PWCHECKDIR)-1, "/pwcheck",
+	    sizeof(fnamebuf) - sizeof(PWCHECKDIR));
+    fnamebuf[MAXPATHLEN-1] = '\0';
 
     (void) unlink(fnamebuf);
 
     memset((char *)&srvaddr, 0, sizeof(srvaddr));
     srvaddr.sun_family = AF_UNIX;
-    strcpy(srvaddr.sun_path, fnamebuf);
+    strncpy(srvaddr.sun_path, fnamebuf, sizeof(srvaddr.sun_path));
     /* Most systems make sockets 0777 no matter what you ask for.
        Known exceptions are Linux and DUX. */
     oldumask = umask((mode_t) 0); /* for Linux, which observes the umask when
 			    setting up the socket */
     r = bind(s, (struct sockaddr *)&srvaddr, sizeof(srvaddr));
     if (r == -1) {
-	perror(fnamebuf);
+	syslog(LOG_ERR, "%.*s: %m",
+	       sizeof(srvaddr.sun_path), srvaddr.sun_path);
 	exit(1);
     }
     umask(oldumask); /* for Linux */
@@ -111,15 +171,15 @@ main()
 				    (harmlessly fails on some systems) */	
     r = listen(s, 5);
     if (r == -1) {
-	perror("listen");
+	syslog(LOG_ERR, "listen: %m");
 	exit(1);
     }
 
     for (;;) {
 	len = sizeof(clientaddr);
 	c = accept(s, (struct sockaddr *)&clientaddr, &len);
-	if (c == -1) {
-	    perror("accept");
+	if (c == -1 && errno != EINTR) {
+	    syslog(LOG_WARNING, "accept: %m");
 	    continue;
 	}
 
@@ -130,7 +190,8 @@ main()
 void newclient(int c)
 {
     char request[1024];
-    int start, n;
+    int n;
+    unsigned int start;
     char *reply;
     extern char *pwcheck();
     
@@ -166,23 +227,26 @@ sendreply:
  * Keep calling the write() system call with 'fd', 'buf', and 'nbyte'
  * until all the data is written out or an error occurs.
  */
-int retry_write(int fd, const char *buf, unsigned nbyte)
+int retry_write(int fd, const char *buf, unsigned int nbyte)
 {
     int n;
     int written = 0;
 
-    if (nbyte == 0) return 0;
+    if (nbyte == 0)
+	return 0;
 
     for (;;) {
         n = write(fd, buf, nbyte);
         if (n == -1) {
-            if (errno == EINTR) continue;
+            if (errno == EINTR)
+		continue;
             return -1;
         }
 
         written += n;
 
-        if (n >= nbyte) return written;
+        if ((unsigned int) n >= nbyte)
+	    return written;
 
         buf += n;
         nbyte -= n;
